@@ -3,9 +3,8 @@
 # This product includes software developed at Datadog (https://www.datadoghq.com/).
 # Copyright 2019 Datadog, Inc.
 
-import base64
 import gzip
-import json
+import ujson as json
 from io import BytesIO, BufferedReader
 from enum import Enum
 from typing import Any
@@ -34,12 +33,13 @@ class EventTypes(_stringTypedEnum):
     CLOUDWATCH_EVENTS = "cloudwatch-events"
     CLOUDFRONT = "cloudfront"
     DYNAMODB = "dynamodb"
+    EVENTBRIDGE = "eventbridge"
     KINESIS = "kinesis"
     LAMBDA_FUNCTION_URL = "lambda-function-url"
     S3 = "s3"
     SNS = "sns"
     SQS = "sqs"
-    EVENTBRIDGE = "eventbridge"
+    STEPFUNCTIONS = "states"
 
 
 class EventSubtypes(_stringTypedEnum):
@@ -109,15 +109,19 @@ def get_first_record(event):
 
 def parse_event_source(event: dict) -> _EventSource:
     """Determines the source of the trigger event"""
-    if type(event) is not dict:
+    if not isinstance(event, dict):
         return _EventSource(EventTypes.UNKNOWN)
 
-    event_source = _EventSource(EventTypes.UNKNOWN)
+    event_source = None
 
+    # Get requestContext safely and ensure it's a dictionary
     request_context = event.get("requestContext")
+    if not isinstance(request_context, dict):
+        request_context = None
+
     if request_context and request_context.get("stage"):
         if "domainName" in request_context and detect_lambda_function_url_domain(
-            request_context.get("domainName")
+            request_context.get("domainName", "")
         ):
             return _EventSource(EventTypes.LAMBDA_FUNCTION_URL)
         event_source = _EventSource(EventTypes.API_GATEWAY)
@@ -125,7 +129,7 @@ def parse_event_source(event: dict) -> _EventSource:
             event_source.subtype = EventSubtypes.API_GATEWAY
         if "routeKey" in event:
             event_source.subtype = EventSubtypes.HTTP_API
-        if event.get("requestContext", {}).get("messageDirection"):
+        if request_context.get("messageDirection"):
             event_source.subtype = EventSubtypes.WEBSOCKET
 
     if request_context and request_context.get("elb"):
@@ -145,12 +149,14 @@ def parse_event_source(event: dict) -> _EventSource:
     if event.get("source") == "aws.events" or has_event_categories:
         event_source = _EventSource(EventTypes.CLOUDWATCH_EVENTS)
 
-    event_record = get_first_record(event)
-    if event_record:
-        aws_event_source = event_record.get(
-            "eventSource", event_record.get("EventSource")
-        )
+    if is_step_function_event(event):
+        event_source = _EventSource(EventTypes.STEPFUNCTIONS)
 
+    event_record = get_first_record(event)
+    if event_record and isinstance(event_record, dict):
+        aws_event_source = event_record.get("eventSource") or event_record.get(
+            "EventSource"
+        )
         if aws_event_source == "aws:dynamodb":
             event_source = _EventSource(EventTypes.DYNAMODB)
         if aws_event_source == "aws:kinesis":
@@ -161,15 +167,16 @@ def parse_event_source(event: dict) -> _EventSource:
             event_source = _EventSource(EventTypes.SNS)
         if aws_event_source == "aws:sqs":
             event_source = _EventSource(EventTypes.SQS)
-
         if event_record.get("cf"):
             event_source = _EventSource(EventTypes.CLOUDFRONT)
 
-    return event_source
+    return event_source or _EventSource(EventTypes.UNKNOWN)
 
 
 def detect_lambda_function_url_domain(domain: str) -> bool:
     #  e.g. "etsn5fibjr.lambda-url.eu-south-1.amazonaws.com"
+    if not isinstance(domain, str):
+        return False
     domain_parts = domain.split(".")
     if len(domain_parts) < 2:
         return False
@@ -189,20 +196,26 @@ def parse_event_source_arn(source: _EventSource, event: dict, context: Any) -> s
     event_record = get_first_record(event)
     # e.g. arn:aws:s3:::lambda-xyz123-abc890
     if source.to_string() == "s3":
-        return event_record.get("s3", {}).get("bucket", {}).get("arn")
+        s3 = event_record.get("s3")
+        if s3:
+            bucket = s3.get("bucket")
+            if bucket:
+                return bucket.get("arn")
+        return None
 
     # e.g. arn:aws:sns:us-east-1:123456789012:sns-lambda
     if source.to_string() == "sns":
-        return event_record.get("Sns", {}).get("TopicArn")
+        sns = event_record.get("Sns")
+        if sns:
+            return sns.get("TopicArn")
+        return None
 
     # e.g. arn:aws:cloudfront::123456789012:distribution/ABC123XYZ
     if source.event_type == EventTypes.CLOUDFRONT:
         distribution_id = (
             event_record.get("cf", {}).get("config", {}).get("distributionId")
         )
-        return "arn:{}:cloudfront::{}:distribution/{}".format(
-            aws_arn, account_id, distribution_id
-        )
+        return f"arn:{aws_arn}:cloudfront::{account_id}:distribution/{distribution_id}"
 
     # e.g. arn:aws:lambda:<region>:<account_id>:url:<function-name>:<function-qualifier>
     if source.equals(EventTypes.LAMBDA_FUNCTION_URL):
@@ -219,31 +232,42 @@ def parse_event_source_arn(source: _EventSource, event: dict, context: Any) -> s
     # e.g. arn:aws:apigateway:us-east-1::/restapis/xyz123/stages/default
     if source.event_type == EventTypes.API_GATEWAY:
         request_context = event.get("requestContext")
-        return "arn:{}:apigateway:{}::/restapis/{}/stages/{}".format(
-            aws_arn, region, request_context.get("apiId"), request_context.get("stage")
-        )
+        api_id = request_context.get("apiId")
+        stage = request_context.get("stage")
+        return f"arn:{aws_arn}:apigateway:{region}::/restapis/{api_id}/stages/{stage}"
 
     # e.g. arn:aws:elasticloadbalancing:us-east-1:123456789012:targetgroup/lambda-xyz/123
     if source.event_type == EventTypes.ALB:
         request_context = event.get("requestContext")
-        return request_context.get("elb", {}).get("targetGroupArn")
+        if request_context:
+            elb = request_context.get("elb")
+            if elb:
+                return elb.get("targetGroupArn")
+        return None
 
     # e.g. arn:aws:logs:us-west-1:123456789012:log-group:/my-log-group-xyz
     if source.event_type == EventTypes.CLOUDWATCH_LOGS:
+        import base64
+
         with gzip.GzipFile(
             fileobj=BytesIO(base64.b64decode(event.get("awslogs", {}).get("data")))
         ) as decompress_stream:
             data = b"".join(BufferedReader(decompress_stream))
         logs = json.loads(data)
         log_group = logs.get("logGroup", "cloudwatch")
-        return "arn:{}:logs:{}:{}:log-group:{}".format(
-            aws_arn, region, account_id, log_group
-        )
+        return f"arn:{aws_arn}:logs:{region}:{account_id}:log-group:{log_group}"
 
     # e.g. arn:aws:events:us-east-1:123456789012:rule/my-schedule
     if source.event_type == EventTypes.CLOUDWATCH_EVENTS and event.get("resources"):
         return event.get("resources")[0]
 
+    # Returning state machine arn as event source arn.
+    if source.event_type == EventTypes.STEPFUNCTIONS:
+        context = event
+        if "_datadog" in event:
+            context = event.get("_datadog")
+        return context.get("StateMachine").get("Id")
+
 
 def get_event_source_arn(source: _EventSource, event: dict, context: Any) -> str:
     event_source_arn = event.get("eventSourceARN") or event.get("eventSourceArn")
@@ -265,17 +289,28 @@ def extract_http_tags(event):
     Extracts HTTP facet tags from the triggering event
     """
     http_tags = {}
+
+    # Safely get request_context and ensure it's a dictionary
     request_context = event.get("requestContext")
+    if not isinstance(request_context, dict):
+        request_context = None
+
     path = event.get("path")
     method = event.get("httpMethod")
+
     if request_context and request_context.get("stage"):
-        if request_context.get("domainName"):
-            http_tags["http.url"] = request_context.get("domainName")
+        domain_name = request_context.get("domainName")
+        if domain_name:
+            http_tags["http.url"] = f"https://{domain_name}"
 
         path = request_context.get("path")
         method = request_context.get("httpMethod")
+
         # Version 2.0 HTTP API Gateway
-        apigateway_v2_http = request_context.get("http")
+        apigateway_v2_http = request_context.get("http", {})
+        if not isinstance(apigateway_v2_http, dict):
+            apigateway_v2_http = {}
+
         if event.get("version") == "2.0" and apigateway_v2_http:
             path = apigateway_v2_http.get("path")
             method = apigateway_v2_http.get("method")
@@ -285,13 +320,34 @@ def extract_http_tags(event):
     if method:
         http_tags["http.method"] = method
 
-    headers = event.get("headers")
+    # Safely get headers
+    headers = event.get("headers", {})
+    if not isinstance(headers, dict):
+        headers = {}
+
     if headers and headers.get("Referer"):
         http_tags["http.referer"] = headers.get("Referer")
 
+    # Try to get `routeKey` from API GW v2; otherwise try to get `resource` from API GW v1
+    route = event.get("routeKey") or event.get("resource")
+    if route and isinstance(route, str):
+        try:
+            # "GET /my/endpoint" = > "/my/endpoint"
+            http_tags["http.route"] = route.split(" ")[-1]
+        except Exception:
+            # If splitting fails, use the route as is
+            http_tags["http.route"] = route
+
     return http_tags
 
 
+_http_event_types = (
+    EventTypes.API_GATEWAY,
+    EventTypes.ALB,
+    EventTypes.LAMBDA_FUNCTION_URL,
+)
+
+
 def extract_trigger_tags(event: dict, context: Any) -> dict:
     """
     Parses the trigger event object to get tags to be added to the span metadata
@@ -305,16 +361,15 @@ def extract_trigger_tags(event: dict, context: Any) -> dict:
         if event_source_arn:
             trigger_tags["function_trigger.event_source_arn"] = event_source_arn
 
-    if event_source.event_type in [
-        EventTypes.API_GATEWAY,
-        EventTypes.ALB,
-        EventTypes.LAMBDA_FUNCTION_URL,
-    ]:
+    if event_source.event_type in _http_event_types:
         trigger_tags.update(extract_http_tags(event))
 
     return trigger_tags
 
 
+_str_http_triggers = [et.value for et in _http_event_types]
+
+
 def extract_http_status_code_tag(trigger_tags, response):
     """
     If the Lambda was triggered by API Gateway, Lambda Function URL, or ALB,
@@ -325,15 +380,7 @@ def extract_http_status_code_tag(trigger_tags, response):
     str_event_source = trigger_tags.get("function_trigger.event_source")
     # it would be cleaner if each event type was a constant object that
     # knew some properties about itself like this.
-    str_http_triggers = [
-        et.value
-        for et in [
-            EventTypes.API_GATEWAY,
-            EventTypes.LAMBDA_FUNCTION_URL,
-            EventTypes.ALB,
-        ]
-    ]
-    if str_event_source not in str_http_triggers:
+    if str_event_source not in _str_http_triggers:
         return
 
     status_code = "200"
@@ -346,3 +393,29 @@ def extract_http_status_code_tag(trigger_tags, response):
         status_code = response.status_code
 
     return str(status_code)
+
+
+def is_step_function_event(event):
+    """
+    Check if the event is a step function that invoked the current lambda.
+
+    The whole event can be wrapped in "Payload" in Legacy Lambda cases. There may also be a
+    "_datadog" for JSONata style context propagation.
+
+    The actual event must contain "Execution", "StateMachine", and "State" fields.
+    """
+    event = event.get("Payload", event)
+
+    # JSONPath style
+    if "Execution" in event and "StateMachine" in event and "State" in event:
+        return True
+
+    # JSONata style
+    dd_context = event.get("_datadog")
+    return (
+        dd_context
+        and "Execution" in dd_context
+        and "StateMachine" in dd_context
+        and "State" in dd_context
+        and "serverless-version" in dd_context
+    )
